/*********************************************/
/*         LCD Device Driver Program         */
/*                                           */
/*                  developed by Eri.Taguchi */
/*                     ver 1.0  Oct. 31 2000 */
/*                     ver 2.0  Nov. 20 2000 */
/*                     ver 2.1  Dec. 14 2000 */
/*********************************************/

#define __KERNEL__
#define MODULE

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/ioctl.h>
#include <linux/sched.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/uaccess.h>

#include "lcd_nr.h"

static char msg[LEN] = {};

static unsigned int busy = 0,in = 0;



static int lcd_open(struct inode *inode, struct file *filp){
  if (busy) return - EBUSY;
  
  printk("lcd_open:\n");
  printk("   filp->f_version     : %lu \n",filp->f_version);

  linecount = LINECOUNT;

  printk("   LINECOUNT: %d \n",linecount);

  busy = 1;
  MOD_INC_USE_COUNT;

  lcd_detect(LCD_PORT_SETUP,LCD_PORT_RDWR);

  if(in==0){
     lcd_module_set(0x0f);
     printk("   lcd_module_set: %x\n",0x0f);
     printk("   repeat        : %x\n",0x0f);
     lcd_module_set(0x38);
     printk("   lcd_module_set: %x\n",0x38);
     printk("   repeat        : %x\n",0x38);
     lcd_module_set(0x01);
     printk("   lcd_module_set: %x\n",0x01);
     printk("   repeat        : %x\n",0x01);
     printk("   repeat        : %x\n",0x01);

     in=1;
  }

  return 0;

}


static int lcd_release(struct inode *inode, struct file *filp){
  busy = 0;
  MOD_DEC_USE_COUNT;

  printk("lcd_close:\n");
  printk("   filp->f_version     : %lu \n",filp->f_version);

  lcd_module_release(LCD_PORT_SETUP,LCD_PORT_RDWR);

  return 0;
}


static loff_t lcd_llseek (struct file *filp, loff_t offset, int origin){
  return - ESPIPE;
}


static ssize_t lcd_read(struct file *file, char *buf, size_t count, loff_t *offset){
  char *str;
  int len = 0;

  str = msg;


  if (linecount == 0){
     return 0;
  }else{
     linecount--;
  }


  *str = *lcd_module_read();

  while(*str != '\0'){
     len++;
     str++;
  }
  str = str - len;
  len= len + 1;
  if (0 == len){
     return -ENOMSG;
  }


  copy_to_user(buf,str,len);

  printk("lcd_read:\n");
  printk("   file->f_version     : %lu \n",file->f_version);
  printk("   requested: %d bytes   returned: %d bytes\n",count,len);
  printk("   message  : %s\n",str);
  return len;
}


static ssize_t lcd_write(struct file *filp, const char *buf, size_t count, loff_t *offset){ 
  char *str;
  int len;

  str = msg;

  if (count < 0){
     return -EINVAL;
  }

  if (LEN - 1 <= count){
     printk("lcd_write_error:\n");
     printk("   offered: %d bytes   obtained: %d bytes\n",count,LEN);
     printk("   message: oversized! \n");

     return -EMSGSIZE;
  }else{
     len = count;

     copy_from_user(str,buf,len);
     lcd_module_write(str,len);
     str[len] = '\0';
  
     printk("lcd_write:\n");
     printk("   offered: %d bytes   obtained: %d bytes\n",count,len);
     printk("   message: '%s'\n",str);

  }

  return len;
}


static int lcd_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg){
  int code,lcd_setcode,i;
  int buf[3];

  printk("lcd_ioctl:\n");

  for(i=0;i<3;i++){
     get_user(buf[i],(int *)(arg+(4*i)));
     printk("buf = %x\n",buf[i]);
  }

  switch(cmd){
     #ifdef LCD_DEBUG
        case LCD_HARDRESET:
	   mod_use_count_ = 1;
     #endif

     case LCD_CLEAR:
	lcd_module_set(0x01);
	printk("   case LCD_CLEAR :\n");
	break;

     case LCD_ADD_ZERO:
	lcd_module_set(0x02);
	printk("   case LCD_ADD_ZERO :\n");
	break;

     case LCD_CURSOR_SHIFT:
	code = 0x04 + 0x02*buf[0] + buf[1];
        lcd_setcode = 16;
	lcd_module_set(code);
	printk("   case LCD_CURSOR_SHIFT   :\n");
        printk("      code =               :%x\n",code);
	break;

     case LCD_ON_OFF:
	code = 0x08 + buf[0];
        lcd_setcode = 16;
	lcd_module_set(code);
	printk("   case LCD_ON_OFF         :\n");
        printk("      code =               :%x\n",code);
	break;

     case LCD_LIST_OR_CURSOR:
	code = 0x10 + 0x08*buf[0] + 0x04*buf[1];
        lcd_setcode = 16;
	lcd_module_set(code);
	printk("   case LCD_LIST_OR_CURSOR :\n");
        printk("      code =               :%x\n",code);
	break;

     case LCD_TWO_SET:
	lcd_module_set(0xc0);
	printk("   case LCD_TWO_SET :\n");
	break;

     case LCD_ADD_SET:
	code = 0x80 + buf[0];
        lcd_setcode = 16;
	lcd_module_set(code);
	break;

     case LCD_BF_ADD:
	lcd_setcode = lcd_module_search();
	put_user(lcd_setcode, (int *)arg);
	break;

     default:
	return -EINVAL;
  }

  return 0;
}


static struct file_operations lcd_fops = {
  lcd_llseek,
  lcd_read,
  lcd_write,
  NULL,                           /* lcd_readdir */
  NULL,                           /* lcd_poll */
  lcd_ioctl,
  NULL,                           /* lcd_mmap */
  lcd_open,
  NULL,                           /* lcd_flush */
  lcd_release,
  NULL,                           /* lcd_fsync */
  NULL,                           /* lcd_fasync */
  NULL,                           /* check_media_change */
  NULL,                           /* revalidate */
  NULL,                           /* lock */
};


static int lcd_detect(unsigned int port, unsigned int range){
  int err;

  printk("lcd_detect:\n");

  if ((err=check_region(port,range)) < 0 ){
    printk("   This ports <%x - %x> are using now!\n",port,range);
    printk("   check_region(port,range) : %d\n",err);
    /* return err; */
  }     /* using port now */
  
  request_region(port,range,"lcd");     /* always success */

  printk("   I/Oport from: %x   to: %x \n",port,range);

  return 0;

}


static void lcd_module_set(int code){
  outb(code,LCD_PORT_SETUP);
}

static int lcd_module_search(void){
  int code;
  code = inb(LCD_PORT_SETUP);
  return (code);
}


static char *lcd_module_read(void){
  int cnt,code1 = CODE1,code2 = CODE2,n=0;
  char *str;

  printk("lcd_module_read:\n");

  str = msg;

  cnt = inb(LCD_PORT_SETUP);

  if(cnt >= CODE1){
     str = "Internal OP is being performed.\n";
     printk("   the condition of operation : %s\n",str);
     printk("   inb(LCD_PORT_SETUP) ->     : %x \n",cnt);
     return (str);
  }

  if(cnt >= 0x40){
     cnt = cnt - 0x40;
     while(cnt > 0){
        lcd_module_set(code2);
        code2 += 0x01;
        printk("2line's character\n");
        printk("2line's character\n");
        printk("2line's character\n");
        str[n] = inb(LCD_PORT_RDWR);
        n++;
        cnt--;
     }
  }else{
     while(cnt > 0){
        lcd_module_set(code1);
        code1 += 0x01;
        printk("1line's character\n");
        printk("1line's character\n");
        printk("1line's character\n");
        str[n] = inb(LCD_PORT_RDWR);
        n++;
        cnt--;
     }
  }
  str[n]='\0';
     printk("   massage                     : '%s'\n",str);


  return(str);
}


static int lcd_module_write(char *str,int len){
  int cnt,cnt2=0,CNT,CNT2;

  printk("lcd_module_write:\n");

  cnt = inb(LCD_PORT_SETUP);
  if(cnt > CODE1){
     printk("   Internal OP is being performed.\n");
     printk("   inb(LCD_PORT_SETUP) ->    : %x \n",cnt);
     return -EAGAIN;  
  }

  while(len > 0){
     if(*str=='^')break;
     CNT = cnt + CODE1;
     lcd_module_set(CNT);
     printk("   module address              : '%x'\n",cnt);
     printk("   module address command      : '%x'\n",CNT);
     outb(*str,LCD_PORT_RDWR);
     printk("   massage                     : '%x'\n",*str);
     cnt += 0x01;
     str++;
     len--;
  }
  if(*str=='^'){
     while(len > 1){
	CNT2 = cnt2 + CODE2; 
	lcd_module_set(CNT2);
        printk("   module address              : '%x'\n",CODE2);
        printk("   module address command      : '%x'\n",CNT2);
        cnt2 += 0x01;
	str++;
	len--;
        outb(*str,LCD_PORT_RDWR);
        printk("   massage                     : '%x'\n",*str);
     }
     return 0;
  }
  return 0;
}


static void lcd_module_release(unsigned int port,unsigned int range){
  release_region(port,range);
  printk("lcd_module_release:\n");
  printk("   I/Oport from: %x   to: %x \n",port,range);
}


int init_module (void){
  int i;

  printk("install 'lcd' into major %d\n",LCD_MAJOR);
  i = register_chrdev (LCD_MAJOR, "lcd", &lcd_fops);

  if (i){
     printk("device registration error\n");
     return -EBUSY;
  }

  return 0;
} 

void cleanup_module (void){
  int i;

  printk("remove 'lcd' from major %d\n",LCD_MAJOR);
  i = unregister_chrdev (LCD_MAJOR,"lcd");

  if (i){
     printk ("unregister_chrdev failed\n");
  }

}
